@@ -281,8 +281,7 @@ int cmd_list_journal(int argc, char *argv[])
 	opt_set(opts, nochanges,	true);
 	opt_set(opts, norecovery,	true);
 	opt_set(opts, read_only,	true);
-	opt_set(opts, degraded,		true);
-	opt_set(opts, very_degraded,		true);
+	opt_set(opts, degraded,		BCH_DEGRADED_very);
 	opt_set(opts, errors,		BCH_ON_ERROR_continue);
 	opt_set(opts, fix_errors,	FSCK_FIX_yes);
 	opt_set(opts, retain_recovery_info ,true);
