@@ -125,7 +125,7 @@ static ranges reserve_new_fs_space(const char *file_path, unsigned block_size,
 	*bcachefs_inum = statbuf.st_ino;
 
 	if (fallocate(fd, 0, 0, size))
-		die("Error reserving space for bcachefs metadata: %m");
+		die("Error reserving space (%llu bytes) for bcachefs metadata: %m", size);
 
 	fsync(fd);
 
@@ -147,7 +147,7 @@ static ranges reserve_new_fs_space(const char *file_path, unsigned block_size,
 		range_add(&extents, e.fe_physical, e.fe_length);
 	}
 	fiemap_iter_exit(&iter);
-	close(fd);
+	xclose(fd);
 
 	ranges_sort_merge(&extents);
 	return extents;
@@ -238,7 +238,7 @@ static int migrate_fs(const char		*fs_path,
 	u64 bcachefs_inum;
 	ranges extents = reserve_new_fs_space(file_path,
 				fs_opts.block_size >> 9,
-				get_size(dev->bdev->bd_fd) / 5,
+				get_size(dev->bdev->bd_fd) / 10,
 				&bcachefs_inum, stat.st_dev, force);
 
 	find_superblock_space(extents, format_opts, dev);
@@ -281,8 +281,10 @@ static int migrate_fs(const char		*fs_path,
 		.type		= BCH_MIGRATE_migrate,
 	};
 
-	u64 reserve_start = round_up((format_opts.superblock_size * 2 + 8) << 9,
-				     dev->opts.bucket_size);
+
+	u64 reserve_start = roundup((format_opts.superblock_size * 2 + 8) << 9,
+				    bucket_bytes(c->devs[0]));
+	BUG_ON(!reserve_start);
 
 	copy_fs(c, fs_fd, fs_path, &s, reserve_start);
 
@@ -425,7 +427,7 @@ int cmd_migrate_superblock(int argc, char *argv[])
 	xpwrite(fd, zeroes, BCH_SB_SECTOR << 9, 0, "zeroing start of disk");
 
 	bch2_super_write(fd, sb);
-	close(fd);
+	xclose(fd);
 
 	/* mark new superblocks */
 
