@@ -115,7 +115,7 @@ int cmd_set_option(int argc, char *argv[])
 				fprintf(stderr, "Can't set option %s\n", opt->attr.name);
 
 			if (opt->flags & OPT_FS) {
-				ret = bch2_opt_check_may_set(c, NULL, i, v);
+				ret = bch2_opt_hook_pre_set(c, NULL, i, v);
 				if (ret < 0) {
 					fprintf(stderr, "error setting %s: %i\n", opt->attr.name, ret);
 					continue;
@@ -133,7 +133,7 @@ int cmd_set_option(int argc, char *argv[])
 							continue;
 						}
 
-						ret = bch2_opt_check_may_set(c, ca, i, v);
+						ret = bch2_opt_hook_pre_set(c, ca, i, v);
 						if (ret < 0) {
 							fprintf(stderr, "error setting %s: %i\n", opt->attr.name, ret);
 							continue;
